package db.service;

import db.Globals;
import db.entity.LogFileMeta;
import db.entity.LogItem;
import db.util.FileUtils;
import lombok.Data;
import lombok.extern.slf4j.Slf4j;
import db.mapper.LogFileMetaMapper;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

@Service
@Data
@Slf4j
public class LogFileService {

    @Autowired
    private LogFileMetaMapper logFileMetaMapper;

    @Value("${system.files.prefix}")
    private String prefix;
    @Value("${system.files.suffix}")
    private String suffix;

    private LogFileMeta logFileMeta; // 当前应写入的文件头信息

    public byte[] queryLog(long loc) throws IOException {
        log.debug("读取{}", loc);
        int fileIndex = (int) (loc % Globals.MAX_LOG_ITEM);
        long fileId = loc / Globals.MAX_LOG_ITEM;

        // 读取fileId文件的头
        LogFileMeta readLogFile = logFileMetaMapper.getLogFileMetaById(fileId);

        byte[] headerContent = new byte[Globals.MAX_LOG_ITEM * 4 + 28];
        FileUtils.readFile(readLogFile.getFileName(), headerContent, 0);
        readLogFile.setPayload(headerContent);
        int fileOffset = readLogFile.getIndices()[fileIndex];

        // 读取logItem的header
        byte[] itemHeaderContent = new byte[16];
        FileUtils.readFile(readLogFile.getFileName(), itemHeaderContent, fileOffset);
        LogItem logItem = new LogItem();
        logItem.setHeader(itemHeaderContent);
        // 读取logItem的body
        byte[] itemContent = new byte[logItem.getLength() - 16];
        FileUtils.readFile(readLogFile.getFileName(), itemContent, fileOffset + 16);
        return itemContent;
    }

    public List<byte[]> queryLogBetweenTime(LogFileMeta logFileMeta, long startTime, long endTime) throws IOException {
        List<byte[]> res = new ArrayList<>();

        // 读取fileId文件的头
        LogFileMeta readLogFile = logFileMetaMapper.getLogFileMetaById(logFileMeta.getId());

        byte[] headerContent = new byte[Globals.MAX_LOG_ITEM * 4 + 28];
        FileUtils.readFile(readLogFile.getFileName(), headerContent, 0);
        readLogFile.setPayload(headerContent);

        // 遍历索引
        for (int i = 0; i < logFileMeta.getCount(); i++) {
            int fileOffset = readLogFile.getIndices()[i];
            // 读取logItem的header
            byte[] itemHeaderContent = new byte[16];
            FileUtils.readFile(readLogFile.getFileName(), itemHeaderContent, fileOffset);
            LogItem logItem = new LogItem();
            logItem.setHeader(itemHeaderContent);
            if (logItem.getStamp() >= startTime && logItem.getStamp() <= endTime) {
                // 读取logItem的body
                byte[] itemContent = new byte[logItem.getLength() - 16];
                FileUtils.readFile(readLogFile.getFileName(), itemContent, fileOffset + 16);
                res.add(itemContent);
            }
        }
        return res;
    }

    public int appendLog(LogItem logItem) throws IOException {
        // 查询db，确认存放文件名
        LogFileMeta lastLogFileMeta = logFileMetaMapper.getLastLogFileMeta();
        if (lastLogFileMeta == null || lastLogFileMeta.getCount() >= Globals.MAX_LOG_ITEM) {
            // 文件已满，创建新文件
            int fileId;
            if (lastLogFileMeta == null) {
                fileId = 0;
            } else {
                fileId = lastLogFileMeta.getId() + 1;
            }
            String fileName = String.format("%s%d%s", prefix, fileId, suffix);
            log.info("旧log文件fileId:{}已满，创建新log文件fileId:{} fileName:{}", fileId-1,
                    fileId, fileName);
            logFileMeta = new LogFileMeta(fileId, fileName);
            logFileMetaMapper.insert(logFileMeta);
            createNewFile();
        } else {
            // 文件没满
            logFileMeta = lastLogFileMeta;
        }

        // 插入磁盘
        int fileIndex = appendLogToDisk(logItem);
        logFileMetaMapper.update(logFileMeta);
        return logFileMeta.getId() * Globals.MAX_LOG_ITEM + fileIndex;
    }

    public int appendLogToDisk(LogItem logItem) throws IOException {

        // 读取文件头
        byte[] headerContent = new byte[Globals.MAX_LOG_ITEM * 4 + 28];
        FileUtils.readFile(logFileMeta.getFileName(), headerContent, 0);
        logFileMeta.setPayload(headerContent);
//        log.debug("读取到log头{}", logFileHeader);

        // 寻找写入位置
        // 读取上一条记录的长度
        int offset = -1;
        Integer index = logFileMeta.getCount();
        if (index == 0) {
            // 第一条
            offset = headerContent.length;
            log.debug("log文件第一条，offset={}", offset);
        } else if (index == Globals.MAX_LOG_ITEM) {
            // 已满
            log.error("log文件最后一条，不应该到这里");
        } else {
            byte[] lastContent = new byte[16];
            FileUtils.readFile(logFileMeta.getFileName(), lastContent,
                    logFileMeta.getIndices()[index - 1]);
            LogItem lastItem = new LogItem();
            lastItem.setHeader(lastContent);
            offset = lastItem.getLength() + logFileMeta.getIndices()[index - 1];
            log.debug("log文件非首尾条，offset={}", offset);
        }

        // 写入数据
        byte[] logItemPayload = logItem.getPayload();
        FileUtils.appendFile(logFileMeta.getFileName(), logItemPayload);

        // 修改文件头
        logFileMeta.getIndices()[index] = offset;
        logFileMeta.setBodyLength(logFileMeta.getBodyLength() + logItemPayload.length);
        if (index == 0) {
            logFileMeta.setMinTime(logItem.getStamp());
            logFileMeta.setMaxTime(logItem.getStamp());
        } else {
            logFileMeta.setMinTime(Math.min(logFileMeta.getMinTime(), logItem.getStamp()));
            logFileMeta.setMaxTime(Math.max(logFileMeta.getMaxTime(), logItem.getStamp()));
        }
        logFileMeta.setCount(index + 1);
        log.debug("修改文件头索引，将{}位置的索引修改为{}", index, offset);
        FileUtils.overwriteFile(logFileMeta.getFileName(), logFileMeta.getPayload(), 0);
        return index;
    }

    public void createNewFile() throws IOException {
        // 创建索引部分
        FileUtils.writeFile(logFileMeta.getFileName(), logFileMeta.getPayload());
    }
}
