package db.handler;

import db.DBApiEntry;
import db.Point;
import db.PointVals;
import db.entity.Probe;
import db.entity.packages.DataResponsePackage;
import db.entity.LogItem;
import db.entity.packages.ListenResponsePackage;
import db.entity.packages.ScanResponsePackage;
import db.service.ProbeService;
import db.service.StatisticService;
import db.util.SpentTimeCalculator;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;
import db.service.LogFileService;

import java.io.IOException;
import java.util.HashMap;

@Component
@Slf4j
public class PackageHandler {

    @Value("${reso-db.metric-name}")
    private String metricName;

    @Autowired
    private DBApiEntry entry;

    @Autowired
    private LogFileService packetFileMetaService;

    @Autowired
    private ProbeService probeService;

    @Autowired
    private StatisticService statisticService;

    public void handleDataResponsePackage(DataResponsePackage dataResponsePackage) throws IOException {
        SpentTimeCalculator spentTimeCalculator = SpentTimeCalculator.create().begin();
        log.debug("开始处理探针上传的数据包:{}", dataResponsePackage);
        statisticService.incrDataCount();

        log.debug("开始保存至磁盘");
        // 封装 LogItem
        LogItem logItem = new LogItem((short) 0x55, dataResponsePackage.getTimestamp(), dataResponsePackage.getData());

        long longIndex;
        try {
            longIndex = packetFileMetaService.appendLog(logItem);
            spentTimeCalculator.end();
        } catch (Exception e) {
            e.printStackTrace();
            log.error("写入数据到磁盘失败");
            log.error("packetFileMetaService.appendLog error", e);
            return;
        }
        statisticService.increDataCountDiskSuccess();
        double index = Double.longBitsToDouble(longIndex);
        log.debug("写入磁盘成功，接收探针上传数据到写入磁盘花费总时间:{}ms 插入{}条", spentTimeCalculator.getSpendTime());

        // 写入数据库
        log.debug("开始尝试插入ResoDB，longIndex={}，doubleIndex={}", longIndex, index);

        HashMap<String, String> tags = new HashMap<>();
        tags.put("sMac", dataResponsePackage.getSMac());
        tags.put("dMac", dataResponsePackage.getDMac());
        tags.put("sPort", dataResponsePackage.getSPort());
        tags.put("dPort", dataResponsePackage.getDPort());
        tags.put("sIp", dataResponsePackage.getSIp());
        tags.put("dIp", dataResponsePackage.getDIp());
        tags.put("probeId", String.valueOf(dataResponsePackage.getProbeId()));
        Point point = new Point(metricName, tags);

        long[] utcTimes = new long[1];
        double[] values = new double[1];

        utcTimes[0] = dataResponsePackage.getTimestamp();
        values[0] = index;
        PointVals pointVals = new PointVals(point, 1, utcTimes, values);
        try {
            int insertCount = entry.insertSinglPoint(pointVals);

            spentTimeCalculator.end();

            if(insertCount!=0) {
                statisticService.increDataCountTsdbSuccess();
                log.debug("插入ResoDB结束，point={}，插入{}条", pointVals, insertCount);
                log.debug("接收探针上传数据到插入时序数据库花费总时间:{}ms 插入{}条", spentTimeCalculator.getSpendTime(), insertCount);
            } else {
                log.error("接收探针上传数据到插入时序数据库花费总时间:{}ms 插入失败", spentTimeCalculator.getSpendTime());
            }

            if(statisticService.getShowStatisticAfterInsertData()) {
                log.info("到目前为止的统计信息：{}", statisticService.getStatistic().toString());
            }
        } catch (Exception e) {
            e.printStackTrace();
            log.error(String.format("插入ResoDB失败，ip:%s port:%d exception:%s", entry.getHost(), entry.getPort(), e));

            if(statisticService.getShowStatisticAfterInsertData()) {
                log.info("到目前为止的统计信息：{}", statisticService.getStatistic().toString());
            }

        }
    }

    public void handleScanResponsePackage(ScanResponsePackage scanResponsePackage) {
        log.info("开始处理数据包ScanResponsePackage:{}", scanResponsePackage);

        // 保存至探针信息
        short probeId = scanResponsePackage.getId();
        if(!probeService.getProbes().containsKey(probeId)) {
            probeService.getProbes().put(probeId, new Probe());
        }
        probeService.getProbes().get(probeId).updateFromScanResponsePackage(scanResponsePackage);

        log.info("成功处理数据包{}", scanResponsePackage);
    }

    public void handleListenResponsePackage(ListenResponsePackage packet) {
        log.info("开始处理数据包ListenResponsePackage: {}", packet);
        log.info("成功处理数据包{}", packet);
    }


}
