package db.netty;

import db.entity.packages.BasePackage;
import io.netty.bootstrap.Bootstrap;
import io.netty.channel.*;
import io.netty.channel.nio.NioEventLoopGroup;
import io.netty.channel.socket.nio.NioDatagramChannel;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.annotation.DependsOn;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Component;

import javax.annotation.PostConstruct;
import javax.annotation.PreDestroy;
import java.util.PriorityQueue;

@Component
@DependsOn("springUtils")
@Slf4j
public class NettyServer {

    EventLoopGroup group = new NioEventLoopGroup();

    @Value("${netty.port}")
    private Integer port;
    @Value("${netty.host}")
    private String host;
    @Value("${netty.useSeperateThread}")
    private Boolean useSeperateThread;

    @Value("${logging.showRecvBytes}")
    private Boolean showRecvBytes;

    private Channel channel;

    @Value("${system.retry-stamp}")
    private Integer retryStamp;
    @Value("${system.retry-times}")
    private Integer retryTimes;

    private PriorityQueue<BasePackage> sendingQueue = new PriorityQueue<>((o1, o2) -> (int) (o1.getSendStamp() - o2.getSendStamp()));


    /**
     * 启动Netty Server
     *
     * @throws InterruptedException
     */
    @PostConstruct
    public void start() throws InterruptedException {
        log.info("开始启动 Netty Server，host:{} port: {}", host, port);
        Bootstrap bootstrap = new Bootstrap();
        NettyServerHandler nettyServerHandler = new NettyServerHandler();
        nettyServerHandler.setUseSeperateThread(useSeperateThread);

        Decoder decoder = new Decoder();
        decoder.setShowRecvBytes(this.showRecvBytes);

        bootstrap.group(group)
                // 指定Channel
                .channel(NioDatagramChannel.class)

                .option(ChannelOption.SO_BROADCAST, true)

                .handler(new ChannelInitializer<NioDatagramChannel>() {
                    @Override
                    public void initChannel(final NioDatagramChannel ch) throws Exception {
                        ChannelPipeline p = ch.pipeline();
                        p.addLast(new Encoder(), decoder);
                        p.addLast(nettyServerHandler);
                    }
                });


        ChannelFuture future = bootstrap.bind(host, port).sync();
        if (future.isSuccess()) {
            log.info("成功启动 Netty Server，host:{} port: {}", host, port);
            channel = future.channel();
        } else {
            log.error(String.format("失败启动 Netty Server，host:%s port: %d", host, port));
        }
    }

    @Scheduled(fixedDelay = 100)
    private void checkQueue() {
//        log.debug("开始探测发送队列");
        while (!sendingQueue.isEmpty()) {
            BasePackage basePackage = sendingQueue.peek();
            if (basePackage.getSendStamp() <= System.currentTimeMillis()) {
                // 需要发送
                sendingQueue.poll();
                log.info("正式发送{}", basePackage);
                channel.writeAndFlush(basePackage);
                if (basePackage.getSendTimes() > 1) {
                    // 再次放入队列
                    basePackage.setSendTimes(basePackage.getSendTimes() - 1);
                    basePackage.setSendStamp(System.currentTimeMillis() + retryStamp);
                    send(basePackage);
                    log.info("超时重传机制，重新添加至发送队列{}", basePackage);
                }
            } else {
                // 暂时无待发送
                return;
            }
        }
    }

    public void send(BasePackage basePackage) {
        // 放入发送队列
        if (basePackage.getSendStamp() == null) {
            // 首次发送，添加时间
            basePackage.setSendTimes(1);
            basePackage.setSendStamp(System.currentTimeMillis());
        }
        log.info("添加消息至发送队列{}", basePackage);
        sendingQueue.add(basePackage);
    }

    @PreDestroy
    public void destroy() {
        group.shutdownGracefully();
        log.info("关闭Netty");
    }
}
